﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

using System.Threading;

namespace wbfsgui
{
    public partial class Form1 : Form
    {
        DataSetWiiGames dsWiiGames;
        DataSetWiiGames dsWiiLocal;
        DataSetDrive dsDriveLocal;
        DataSetDrive dsDriveWii;
        FileInfo[] fi;
        DataSetDrive.driveRow wiiHDD;
        DataSetDrive.driveRow wiiLocalHDD;
        bool validHDD;
        bool validLocalWii;
        bool validLocalDVD;

        System.Net.WebClient web = new System.Net.WebClient();
        wbfsCall wc;

        output_progress progressCallback;
        output_text textCallback;

        HiPerfTimer ptProcess = new HiPerfTimer();

        #region Form Management
        public Form1()
        {
            InitializeComponent();

            this.Text = ".: wbfsGUI :: V" + System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString() +" :.";

            progressCallback = new output_progress(updateProgress);
            textCallback = new output_text(outputText);
            WBFS.wbfs_init_output_progress(progressCallback);
            WBFS.wbfs_init_output_text(textCallback);

            string empty = "";
            txtOutput.Text = empty;

            if (!Directory.Exists("./images"))
            {
                Directory.CreateDirectory("./images");
            }

            dsDriveLocal = new DataSetDrive();
            dsDriveWii = new DataSetDrive();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            wc = new wbfsCall(this);

            if (Properties.Settings.Default.LocalPath.Equals("") || !Directory.Exists(Properties.Settings.Default.LocalPath))
            {
                Properties.Settings.Default.LocalPath = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().CodeBase).Substring(6);
            }

            int x = splitContainer1.SplitterRectangle.Right;
            int y = splitContainer1.SplitterRectangle.Top;
            int width = splitContainer1.SplitterRectangle.Width;
            int height = splitContainer1.SplitterRectangle.Height;

            int left = x - btnAllToLocal.Size.Width;
            int top = y + height / 2 - 120 / 2 + btnAllToLocal.Size.Height / 2;

            btnAllToWii.Location = new Point(left, top);
            btnToWii.Location = new Point(left, top + 31);
            btnToLocal.Location = new Point(left, top + 62);
            btnAllToLocal.Location = new Point(left, top + 93);

            dsWiiGames = new DataSetWiiGames();
            dsWiiLocal = new DataSetWiiGames();

            listWii.DataSource = dsWiiGames.wiiGames;
            listWii.Columns["ID"].ReadOnly = true;
            listWii.Columns["GameSize"].ReadOnly = true;
            listWii.Columns["Filesize"].Visible = false;
            listWii.Columns["Filename"].Visible = false;
            listWii.Columns["Filepath"].Visible = false;

            listLocal.DataSource = dsWiiLocal.wiiGames;
            listLocal.Columns["Filepath"].Visible = false;
            listWii.Columns["GameName"].ReadOnly = false;

            DetectAllDrive();

            txtWiiHDD.ValueMember = "HandleName";
            txtWiiHDD.DisplayMember = "MountPoint";
            txtWiiHDD.DataSource = dsDriveWii.drive;

            txtBrowse.ValueMember = "HandleName";
            txtBrowse.DisplayMember = "MountPoint";
            txtBrowse.DataSource = dsDriveLocal.drive;

            validHDD = false;
            txtWiiHDD.Text = Properties.Settings.Default.WiiDrive;

            for (int i = 0; i < dsDriveWii.drive.Rows.Count && !validHDD; ++i)
            {
                txtWiiHDD.SelectedIndex = i;
            }

            if (!validHDD)
                txtWiiHDD.SelectedIndex = 0;

            listWii.Resize += new EventHandler(listWii_Resize);

            includeSubdirectoriesToolStripMenuItem.Checked = Properties.Settings.Default.SearchSubDirectories;

            txtBrowse.Text = null;
            txtBrowse.Text = Properties.Settings.Default.LocalPath;

            useLocalImagesOnlyToolStripMenuItem.Checked = !Properties.Settings.Default.DownloadMissingImage;

            ReadGameOnWii();
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            if(Directory.Exists(txtBrowse.Text))
                Properties.Settings.Default.LocalPath = txtBrowse.Text;

            if (validHDD)
                Properties.Settings.Default.WiiDrive = wiiHDD["MountPoint"].ToString();

            Properties.Settings.Default.SearchSubDirectories = includeSubdirectoriesToolStripMenuItem.Checked;
            Properties.Settings.Default.DownloadMissingImage = !useLocalImagesOnlyToolStripMenuItem.Checked;

            Properties.Settings.Default.Save();
        }

        private void Form1_SizeChanged(object sender, EventArgs e)
        {
            int x = splitContainer1.SplitterRectangle.Right;
            int y = splitContainer1.SplitterRectangle.Top;
            int width = splitContainer1.SplitterRectangle.Width;
            int height = splitContainer1.SplitterRectangle.Height;

            int left = x - btnAllToLocal.Size.Width;
            int top = y + height / 2 - 120 / 2 + btnAllToLocal.Size.Height / 2;

            btnAllToWii.Location = new Point(left, top);
            btnToWii.Location = new Point(left, top + 31);
            btnToLocal.Location = new Point(left, top + 62);
            btnAllToLocal.Location = new Point(left, top + 93);
        }
        #endregion

        #region TextBox Management
        public void outputText(string text)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new AppendText(outputText), text);
            }
            else
            {
                txtOutput.AppendText(text+"\r\n");
            }
        }

        private void txtBrowse_TextChanged(object sender, EventArgs e)
        {
            wiiLocalHDD = null;
            validLocalWii = false;
            validLocalDVD = false;

            dsWiiLocal.Clear();
            if ((sender as ComboBox).Text != null)
            {
                string text = (sender as ComboBox).Text;

                if (!text.Equals(""))
                {
                    StringBuilder sb = new StringBuilder("", 0x100 * 4096);

                    if (dsDriveLocal.drive.Select("[MountPoint]='" + text + "'").Length > 0)
                        wiiLocalHDD = dsDriveLocal.drive.Select("[MountPoint]='" + text + "'")[0] as DataSetDrive.driveRow;

                    if (wiiLocalHDD != null)
                        validLocalWii = WBFS.wbfs_applet_list(wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), sb) != -1;
                }

                if (!validLocalWii)
                {
                    if (text.LastIndexOf("\\") + 1 != text.Length)
                    {
                        text += "\\";
                    }

                    if (Directory.Exists(text))
                    {
                        wc.WorkingDirectory = text;
                    }
                }

                ReadLocalGame();
            }
        }

        private void txtWiiHDD_TextChanged(object sender, EventArgs e)
        {
            wiiHDD = null;
            validHDD = false;
            dsWiiGames.Clear();
            if ((sender as ComboBox).SelectedIndex >= 0)
            {
                string text = (sender as ComboBox).Text;

                // Detect Wii HDD //
                wiiHDD = dsDriveWii.drive.Rows[(sender as ComboBox).SelectedIndex] as DataSetDrive.driveRow;

                StringBuilder sb = new StringBuilder("", 0x100 * 4096);

                validHDD = WBFS.wbfs_applet_list(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), sb) != -1;
            }

            ReadGameOnWii();
        }

        private void txtWiiHDD_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsLetter(e.KeyChar) && e.KeyChar >= 32)
                e.Handled = true;
        }
        #endregion

        private delegate void TwoIntDelegate(int x, int y);
        private delegate void AppendText(string text);
        private delegate void NoArgs();

        #region Refresh Function
        public void DetectAllDrive()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(DetectAllDrive));
            }
            else
            {
                int index = txtWiiHDD.SelectedIndex;
                string browse = txtBrowse.Text;

                // Detect Wii HDD //
                const uint N = 4096;
                StringBuilder mounted_drive = new StringBuilder((int)N, (int)N);
                StringBuilder volume = new StringBuilder((int)N, (int)N);
                WBFS.wbfs_applet_list_drive(mounted_drive, volume);

                string[] split_char = { "%#%" };

                string[] drive_list = mounted_drive.ToString().Split(split_char, StringSplitOptions.RemoveEmptyEntries);

                dsDriveLocal.Clear();
                for (int i = 0; i < drive_list.Length; ++i)
                {
                    DataSetDrive.driveRow dr = dsDriveLocal.drive.NewdriveRow();
                    dr.MountPoint = drive_list[i].Substring(drive_list[i].IndexOf("\\\\?\\") + 4, 2);
                    dr.Offset = 0;
                    dr.PhysicalDrive = drive_list[i].Substring(drive_list[i].IndexOf("PhysicalDrive") + 13, drive_list[i].IndexOf("@", drive_list[i].IndexOf("PhysicalDrive")) - drive_list[i].IndexOf("PhysicalDrive") - 13);
                    dr.Type = "";
                    dr.Partition = drive_list[i].Substring(drive_list[i].IndexOf("Part") + 4);
                    dr.HandleName = drive_list[i].Substring(drive_list[i].IndexOf("\\\\?\\"), drive_list[i].IndexOf("@", drive_list[i].IndexOf("\\\\?\\")));

                    if (dsDriveLocal.drive.FindByPhysicalDrivePartition(dr.PhysicalDrive, dr.Partition) == null)
                        dsDriveLocal.drive.AdddriveRow(dr);
                }

                drive_list = volume.ToString().Split(split_char, StringSplitOptions.RemoveEmptyEntries);

                for (int i = 0; i < drive_list.Length; ++i)
                {
                    DataSetDrive.driveRow dr = dsDriveLocal.drive.NewdriveRow();
                    dr.MountPoint = "";
                    dr.Offset = Convert.ToInt32(drive_list[i].Substring(drive_list[i].IndexOf("LBA") + 3, drive_list[i].IndexOf("@", drive_list[i].IndexOf("LBA")) - drive_list[i].IndexOf("LBA") - 3));
                    dr.PhysicalDrive = drive_list[i].Substring(drive_list[i].IndexOf("PhysicalDrive") + 13, drive_list[i].IndexOf("@", drive_list[i].IndexOf("PhysicalDrive")) - drive_list[i].IndexOf("PhysicalDrive") - 13);
                    dr.Type = drive_list[i].Substring(0, drive_list[i].IndexOf("@"));
                    dr.Partition = drive_list[i].Substring(drive_list[i].IndexOf("Part") + 4, drive_list[i].IndexOf("@", drive_list[i].IndexOf("Part")) - drive_list[i].IndexOf("Part") - 4);
                    dr.HandleName = drive_list[i].Substring(drive_list[i].IndexOf("\\\\.\\"), drive_list[i].IndexOf("@", drive_list[i].IndexOf("\\\\.\\")) - drive_list[i].IndexOf("\\\\.\\"));

                    dr.MountPoint = dr.Type + " @ PhysicalDrive" + dr.PhysicalDrive + "Partition" + dr.Partition;

                    if (dsDriveLocal.drive.FindByPhysicalDrivePartition(dr.PhysicalDrive, dr.Partition) == null)
                        dsDriveLocal.drive.AdddriveRow(dr);
                }

                dsDriveWii = dsDriveLocal.Copy() as DataSetDrive;

                if(index >= 0)
                    txtWiiHDD.SelectedIndex = index;

                if (!browse.Equals(""))
                {
                    txtBrowse.Text = null;
                    txtBrowse.Text = browse;
                }
            }
        }

        public void ReadGameOnWii()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(ReadGameOnWii));
            }
            else
            {
                lblUsedSpace.Text = "Unable to Read Info";
                dsWiiGames.Clear();

                if (validHDD)
                {
                    ulong used = 0;
                    ulong free = 0;
                    ulong total = 0;

                    WBFS.wbfs_applet_info(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), ref used, ref free, ref total);

                    if (total != 0)
                    {
                        progressBar1.Value = Convert.ToInt32(100 * used / total);
                        lblUsedSpace.Text = String.Format("{0:0.00} / {1:0.00} GB", used / 1024.0 / 1024 / 1024, total / 1024.0 / 1024 / 1024);
                    }

                    StringBuilder sb = new StringBuilder("", 0x100 * 4096);
                    WBFS.wbfs_applet_list(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), sb);

                    string[] split_char = { "%#%" };
                    string[] disc_list = sb.ToString().Split(split_char, StringSplitOptions.RemoveEmptyEntries);

                    int limit = disc_list.Length;
                    for (int j = 0; j < limit; ++j)
                    {
                        if (disc_list[j].Trim().Length == 0)
                            continue;

                        string wiiGame = disc_list[j];

                        DataSetWiiGames.wiiGamesRow dr = dsWiiGames.wiiGames.NewwiiGamesRow();
                        dr.ID = wiiGame.Substring(0, 6);

                        wiiGame = wiiGame.Substring(7);
                        wiiGame = wiiGame.Trim();

                        dr.GameSize = wiiGame.Substring(0, wiiGame.IndexOf(" "));
                        dr.GameName = wiiGame.Substring(wiiGame.IndexOf(" ") + 1);
                        dr.GameName = dr.GameName.Trim();

                        dsWiiGames.wiiGames.AddwiiGamesRow(dr);
                    }
                }

                lblUsedSpace.Location = new Point(progressBar1.Location.X + progressBar1.Size.Width - lblUsedSpace.Size.Width, lblUsedSpace.Location.Y);
            }
        }

        public void ReadLocalGame()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(ReadLocalGame));
            }
            else
            {
                dsWiiLocal.Clear();
                lblUsedSpaceLocal.Text = "Unable to Read Info";

                if (validLocalWii)
                {
                    StringBuilder sb = new StringBuilder("", 0x100 * 4096);
                    WBFS.wbfs_applet_list(wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), sb);

                    string[] split_char = { "%#%" };

                    string[] disc_list = sb.ToString().Split(split_char, StringSplitOptions.RemoveEmptyEntries);
                    int limit = disc_list.Length;
                    for (int j = 0; j < limit; ++j)
                    {
                        if (disc_list[j].Trim().Length == 0)
                            continue;

                        string wiiGame = disc_list[j].ToString();
                        DataSetWiiGames.wiiGamesRow dr = dsWiiLocal.wiiGames.NewwiiGamesRow();
                        dr.ID = wiiGame.Substring(0, 6);

                        wiiGame = wiiGame.Substring(7);
                        wiiGame = wiiGame.Trim();

                        dr.GameSize = wiiGame.Substring(0, wiiGame.IndexOf(" "));
                        dr.GameName = wiiGame.Substring(wiiGame.IndexOf(" ") + 1);
                        dr.GameName = dr.GameName.Trim();

                        dsWiiLocal.wiiGames.AddwiiGamesRow(dr);
                    }

                    ulong used = 0;
                    ulong free = 0;
                    ulong total = 0;

                    WBFS.wbfs_applet_info(wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), ref used, ref free, ref total);

                    if (total != 0)
                    {
                        progressBar2.Value = Convert.ToInt32(100 * used / total);
                        progressBar2.Maximum = 100;
                        lblUsedSpaceLocal.Text = String.Format("{0:0.00} / {1:0.00} GB  (WBFS)", used / 1024.0 / 1024 / 1024, total / 1024.0 / 1024 / 1024);
                    }
                }
                else if (txtBrowse.Text != null)
                {
                    string text = txtBrowse.Text;

                    if (text.Length > 0)
                    {
                        if (Char.IsLetter(text[0]) && (text.Length == 1 || (text.Length > 1 && text[1].Equals(':'))))
                        {
                            DriveInfo dr = new DriveInfo(text[0].ToString());

                            if (dr.IsReady)
                            {
                                progressBar2.Maximum = Convert.ToInt32(dr.TotalSize / 1024 / 1024 * 100 / 1024);
                                progressBar2.Value = Convert.ToInt32((dr.TotalSize - dr.TotalFreeSpace) / 1024 / 1024 * 100 / 1024);

                                lblUsedSpaceLocal.Text = (progressBar2.Value / 100.0f).ToString() + " / " + (progressBar2.Maximum / 100.0f).ToString() + "GB  (Local)";
                            }

                            if (dr.DriveType == DriveType.CDRom)
                            {
                                validLocalDVD = true;
                            }
                        }
                    }

                    if (Directory.Exists(text))
                    {
                        DirectoryInfo di = new DirectoryInfo(text);
                        fi = di.GetFiles("*.iso");

                        for (int i = 0; i < fi.Length; ++i)
                        {
                            FileStream fs = File.OpenRead(fi[i].FullName);
                            byte[] arrayFile = new byte[1 << 10];
                            fs.Read(arrayFile, 0, 1 << 10);
                            Encoding enc = Encoding.ASCII;
                            string header = enc.GetString(arrayFile);
                            fs.Close();

                            uint result = WBFS.wbfs_applet_estimate(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), fi[i].FullName);

                            DataSetWiiGames.wiiGamesRow dr = dsWiiLocal.wiiGames.NewwiiGamesRow();
                            dr.ID = header.Substring(0, 6).Replace("\0", "");
                            dr.GameName = header.Substring(32, 64).Replace("\0", "");
                            dr.Filename = fi[i].Name;
                            dr.Filepath = fi[i].FullName;
                            dr.Filesize = String.Format("{0:0.00} GB", fi[i].Length / 1024.0 / 1024 / 1024);

                            if (result > 0)
                            {
                                dr.GameSize = String.Format("{0:0.00} GB", result / 1024.0 / 1024 / 1024);
                            }

                            if (dr.GameName.Length > 0 && dr.ID.Length == 6)
                                dsWiiLocal.wiiGames.AddwiiGamesRow(dr);
                        }

                        if (includeSubdirectoriesToolStripMenuItem.Checked)
                            AddSubDirectoryGames(di);
                    }
                }
            }
        }

        public void AddSubDirectoryGames(DirectoryInfo di)
        {
            foreach (DirectoryInfo dis in di.GetDirectories())
            {
                try
                {
                    fi = dis.GetFiles("*.iso");
                }
                catch
                {
                    continue;
                }

                for (int i = 0; i < fi.Length; ++i)
                {
                    FileStream fs = File.OpenRead(fi[i].FullName);
                    byte[] arrayFile = new byte[1 << 10];
                    fs.Read(arrayFile, 0, 1 << 10);
                    Encoding enc = Encoding.ASCII;
                    string header = enc.GetString(arrayFile);
                    fs.Close();

                    uint result = WBFS.wbfs_applet_estimate(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), fi[i].FullName);

                    DataSetWiiGames.wiiGamesRow dr = dsWiiLocal.wiiGames.NewwiiGamesRow();
                    dr.ID = header.Substring(0, 6).Replace("\0", "");
                    dr.GameName = header.Substring(32, 64).Replace("\0", "");
                    dr.Filename = fi[i].Name;
                    dr.Filepath = fi[i].FullName;
                    dr.Filesize = String.Format("{0:0.00} GB", fi[i].Length / 1024.0 / 1024 / 1024);

                    if (result > 0)
                    {
                        dr.GameSize = String.Format("{0:0.00} GB", result / 1024.0 / 1024 / 1024);
                    }

                    if (dr.GameName.Length > 0 && dr.ID.Length == 6)
                        dsWiiLocal.wiiGames.AddwiiGamesRow(dr);
                }

                AddSubDirectoryGames(dis);
            }
        }

        private void SetGameID(PictureBox pb, string gameID)
        {
            if (!useLocalImagesOnlyToolStripMenuItem.Checked && !File.Exists("./images/" + gameID + ".png"))
            {
                try
                {
                    web.DownloadFile(new System.Uri("http://www.theotherzone.com/wii/pal/" + gameID + ".png"), "./images/" + gameID + ".png");
                }
                catch (Exception) { }

                if (!File.Exists("./images/" + gameID + ".png"))
                {
                    try
                    {
                        web.DownloadFile(new System.Uri("http://www.theotherzone.com/wii/ntsc/" + gameID + ".png"), "./images/" + gameID + ".png");
                    }
                    catch (Exception) { }
                }

                if (!File.Exists("./images/" + gameID + ".png"))
                {
                    try
                    {
                        web.DownloadFile(new System.Uri("http://www.theotherzone.com/wii/ntscj/" + gameID + ".png"), "./images/" + gameID + ".png");
                    }
                    catch (Exception) { }
                }
            }

            {
                if (!File.Exists("./images/" + gameID + ".png"))
                {
                    gameID = "noimage";
                }

                if (gameID != "noimage")
                {
                    pb.Load("./images/" + gameID + ".png");
                }
                else
                {
                    pb.Image = global::wbfsgui.Properties.Resources.noimage;
                }
            }
        }
        #endregion

        #region Button Management
        private void btnToWii_Click(object sender, EventArgs e)
        {
            if (validHDD && validLocalWii)
            {
                List<String> l = new List<string>();
                foreach (DataGridViewCell dgvc in listLocal.SelectedCells)
                {
                    l.Add(dgvc.OwningRow.Cells["ID"].Value.ToString());
                }

                foreach (DataGridViewRow dgvr in listLocal.SelectedRows)
                {
                    l.Add(dgvr.Cells["ID"].Value.ToString());
                }

                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < l.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_COPY, wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), l[i]);
                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
            else
            {
                List<String> l = new List<string>();
                foreach (DataGridViewCell dgvc in listLocal.SelectedCells)
                {
                    l.Add(dgvc.OwningRow.Cells["Filepath"].Value.ToString());
                }

                foreach (DataGridViewRow dgvr in listLocal.SelectedRows)
                {
                    l.Add(dgvr.Cells["Filepath"].Value.ToString());
                }

                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < l.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_ADD, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, l[i]);
                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }

                if (l.Count == 0 && validLocalDVD)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_ADD, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, txtBrowse.Text);
                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
        }

        private void btnAllToWii_Click(object sender, EventArgs e)
        {
            if (validHDD && validLocalWii)
            {
                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < dsWiiLocal.wiiGames.Rows.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_COPY, wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), dsWiiLocal.wiiGames.Rows[i]["ID"].ToString());

                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
            else
            {
                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < dsWiiLocal.wiiGames.Rows.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_ADD, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, dsWiiLocal.wiiGames.Rows[i]["ID"].ToString());

                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
        }

        private void btnToLocal_Click(object sender, EventArgs e)
        {
            if (validHDD && validLocalWii)
            {
                List<String> l = new List<string>();
                foreach (DataGridViewCell dgvc in listWii.SelectedCells)
                {
                    l.Add(dgvc.OwningRow.Cells["ID"].Value.ToString());
                }

                foreach (DataGridViewRow dgvr in listWii.SelectedRows)
                {
                    l.Add(dgvr.Cells["ID"].Value.ToString());
                }

                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < l.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_COPY, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), l[i]);

                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
            else if (txtBrowse.Text != null)
            {
                string text = txtBrowse.Text;

                if (Directory.Exists(text))
                {
                    List<String> l = new List<string>();
                    foreach (DataGridViewCell dgvc in listWii.SelectedCells)
                    {
                        l.Add(dgvc.OwningRow.Cells["ID"].Value.ToString());
                    }

                    foreach (DataGridViewRow dgvr in listWii.SelectedRows)
                    {
                        l.Add(dgvr.Cells["ID"].Value.ToString());
                    }

                    prgOverall.Value = 0;
                    prgOverall.Maximum = 0;
                    for (int i = 0; i < l.Count; ++i)
                    {
                        wc.AddAction(wbfsCall.eAction.eACTION_EXTRACT, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, l[i]);

                        ++prgOverall.Maximum;
                        ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                    }
                }
            }
        }

        private void btnAllToLocal_Click(object sender, EventArgs e)
        {
            if (validHDD && validLocalWii)
            {
                prgOverall.Value = 0;
                prgOverall.Maximum = 0;
                for (int i = 0; i < dsWiiGames.wiiGames.Rows.Count; ++i)
                {
                    wc.AddAction(wbfsCall.eAction.eACTION_COPY, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), wiiLocalHDD["HandleName"].ToString(), Convert.ToInt32(wiiLocalHDD["Offset"]), dsWiiGames.wiiGames.Rows[i]["ID"].ToString());

                    ++prgOverall.Maximum;
                    ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                }
            }
            else if (txtBrowse.Text != null)
            {
                string text = txtBrowse.Text;

                if (Directory.Exists(text))
                {
                    prgOverall.Value = 0;
                    prgOverall.Maximum = 0;
                    for (int i = 0; i < dsWiiGames.wiiGames.Rows.Count; ++i)
                    {
                        wc.AddAction(wbfsCall.eAction.eACTION_EXTRACT, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, dsWiiGames.wiiGames.Rows[i]["ID"].ToString());

                        ++prgOverall.Maximum;
                        ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
                    }
                }
            }
        }

        private void btnRemove_Click(object sender, EventArgs e)
        {
            List<String> l = new List<string>();
            foreach (DataGridViewCell dgvc in listWii.SelectedCells)
            {
                l.Add(dgvc.OwningRow.Cells["ID"].Value.ToString());
            }

            foreach (DataGridViewRow dgvr in listWii.SelectedRows)
            {
                l.Add(dgvr.Cells["ID"].Value.ToString());
            }

            pgrProgress.Value = 0;
            for (int i = 0; i < l.Count; ++i)
            {
                wc.AddAction(wbfsCall.eAction.eACTION_REMOVE, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, l[i]);

                ++prgOverall.Maximum;
                ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
            }
        }

        private void btnInit_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("All data on drive will be erased", "Warning", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) != DialogResult.Cancel)
            {
                wc.AddAction(wbfsCall.eAction.eACTION_INIT, wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), "", 0, "");

                prgOverall.Value = 0;
                prgOverall.Maximum = 1;
                ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin));
            }
        }

        private void btnBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.SelectedPath = wc.WorkingDirectory;

            fbd.ShowDialog();

            if (fbd.SelectedPath != null && !fbd.SelectedPath.Equals(""))
            {
                txtBrowse.Text = null;
                txtBrowse.Text = fbd.SelectedPath;
            }
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            DisableButtons();
            ReadGameOnWii();
            ReadLocalGame();
            EnableButtons();
        }

        public void EnableButtons()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(EnableButtons));
            }
            else
            {
                btnBrowse.Enabled = true;
                btnRefresh.Enabled = true;
                btnInit.Enabled = true;
                btnAllToLocal.Enabled = true;
                btnAllToWii.Enabled = true;
                btnToLocal.Enabled = true;
                btnToWii.Enabled = true;
                btnRemove.Enabled = true;
            }
        }

        public void DisableButtons()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(DisableButtons));
            }
            else
            {
                btnBrowse.Enabled = false;
                btnRefresh.Enabled = false;
                btnInit.Enabled = false;
                btnAllToLocal.Enabled = false;
                btnAllToWii.Enabled = false;
                btnToLocal.Enabled = false;
                btnToWii.Enabled = false;
                btnRemove.Enabled = false;
            }
        }
        #endregion

        #region Progress Management
        public void updateProgress(int x, int y)
        {
            if (pgrProgress.InvokeRequired)
            {
                pgrProgress.Invoke(new TwoIntDelegate(updateProgress), x, y);
            }
            else
            {

                if (x == 0)
                {
                    ptProcess.Start();
                }

                if (x == y)
                {
                    int d = Convert.ToInt32(ptProcess.Duration);
                    int h = d / 3600;
                    int m = (d / 60) % 60;
                    int s = d % 60;
                    txtOutput.AppendText(String.Format("Done in {0:00}:{1:00}:{2:00}\r\n", h, m, s));
                }

                {
                    double d = ptProcess.Duration;

                    if (x != 0)
                        d = d * y / x - d;

                    int h = Convert.ToInt32(d / 3600 - 0.5);
                    int m = Convert.ToInt32(d / 60 - 0.5) % 60;
                    int s = Convert.ToInt32(d - 0.5) % 60;


                    txtETA.Text = String.Format("{0:00}:{1:00}:{2:00}", h, m, s);
                }


                pgrProgress.Value = x;
                pgrProgress.Maximum = y;
            }
        }

        public void IncrementOverall()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new NoArgs(IncrementOverall));
            }
            else
            {
                ++prgOverall.Value;
            }
        }
        #endregion

        #region Options Management
        private void includeSubdirectoriesToolStripMenuItem_CheckStateChanged(object sender, EventArgs e)
        {
            ReadLocalGame();
        }
        #endregion

        #region List Management
        void listWii_Resize(object sender, EventArgs e)
        {
            listWii.AutoResizeColumns();
        }

        private void listWii_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            WBFS.wbfs_applet_rename(wiiHDD["HandleName"].ToString(), Convert.ToInt32(wiiHDD["Offset"]), listWii.Rows[e.RowIndex].Cells["ID"].Value.ToString(), listWii.Rows[e.RowIndex].Cells["GameName"].Value.ToString());
        }

        private void listLocal_SelectionChanged(object sender, EventArgs e)
        {
            DataGridView dgv = sender as DataGridView;

            if (dgv.SelectedCells.Count > 0)
            {
                SetGameID(dgv.Name.Equals("listLocal") ? pictureBoxLocal : pictureBoxWii, dgv.SelectedCells[0].OwningRow.Cells["ID"].Value.ToString());
            }
            else if (dgv.SelectedRows.Count > 0)
            {
                SetGameID(dgv.Name.Equals("listLocal") ? pictureBoxLocal : pictureBoxWii, dgv.SelectedRows[0].Cells["ID"].Value.ToString());
            }
        }

        private void listLocal_RowsAdded(object sender, DataGridViewRowsAddedEventArgs e)
        {
            /*DataGridViewRow dgvr = listLocal.Rows[e.RowIndex];

            if (dgvr.Cells["ID"].Value.Equals("") || dgvr.Cells["GameName"].Value.Equals(""))
            {
                dgvr.DefaultCellStyle.BackColor = Color.LightCoral;
            }
            else if (dgvr.Cells["GameSize"].Value.Equals(""))
            {
                dgvr.DefaultCellStyle.BackColor = Color.Gold;
            }
            else
            {
                dgvr.DefaultCellStyle.BackColor = Color.GreenYellow;
            }*/
        }
        #endregion

        #region Tools Management
        private void buildHomebrewChannelDirectoriesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*string arguments = wiiHDD + " makehbc";

            prgOverall.Value = 0;
            prgOverall.Maximum = 1;
            ThreadPool.QueueUserWorkItem(new WaitCallback(wc.RunWbfsWin), (object)arguments);*/
        }

        private void donateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Process.Start(@"https://www.paypal.com/cgi-bin/webscr?cmd=_donations&business=JN755QAP96RUE&lc=CA&item_name=wfbsGUI&currency_code=USD&bn=PP-DonationsBF:btn_donate_LG.gif:NonHosted");
        }

        private void websiteToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Process.Start(@"http://duchaine.dyndns.org/index.php/wbfsgui.html");
        }

        private void redetectAllDrivesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DetectAllDrive();
        }
        #endregion
    }

    public class wbfsCall
    {
        private static Mutex mutex = new Mutex();
        Form1 frm;

        class task
        {
            public task(eAction action, string source, int source_offset, string destination, int destination_offset, string args)
            {
                action_ = action;
                source_ = source;
                source_offset_ = source_offset;
                destination_ = destination;
                destination_offset_ = destination_offset;
                args_ = args;
            }

            public eAction action_;
            public string source_;
            public int source_offset_;
            public string destination_;
            public int destination_offset_;
            public string args_;
        };

        private static Mutex tasks_mutex = new Mutex();
        List<task> tasks;

        string workingDirectory;

        public string WorkingDirectory
        {
            get { return workingDirectory; }
            set { workingDirectory = value; }
        }

        public wbfsCall(Form1 f)
        {
            tasks = new List<task>();
            frm = f;
        }

        public enum eAction
        {
            eACTION_INIT,
            eACTION_ADD,
            eACTION_EXTRACT,
            eACTION_COPY,
            eACTION_REMOVE
        }

        public void AddAction(eAction action, string source, int source_offset, string destination, int destination_offset, string args)
        {
            tasks_mutex.WaitOne();

            tasks.Add(new task(action, source, source_offset, destination, destination_offset, args));

            tasks_mutex.ReleaseMutex();
        }

        public void RunWbfsWin(object empty)
        {
            mutex.WaitOne();

            frm.DisableButtons();

            tasks_mutex.WaitOne();
            task t = null;

            if(tasks.Count > 0)
                t = tasks[0];

            tasks_mutex.ReleaseMutex();

            if (t == null)
            {
                mutex.ReleaseMutex();
                return;
            }

            switch (t.action_)
            {
                case eAction.eACTION_REMOVE:
                    {
                        WBFS.wbfs_applet_remove(t.source_, t.source_offset_, t.args_);
                        frm.ReadGameOnWii();
                    }
                    break;

                case eAction.eACTION_ADD:
                    {
                        WBFS.wbfs_applet_add(t.source_, t.source_offset_, t.args_);
                        frm.ReadGameOnWii();
                    }
                    break;

                case eAction.eACTION_EXTRACT:
                    {
                        WBFS.wbfs_applet_extract(t.source_, t.source_offset_, t.args_, WorkingDirectory, "");
                        frm.ReadLocalGame();
                    }
                    break;

                case eAction.eACTION_INIT:
                    {
                        WBFS.wbfs_applet_init(t.source_, t.source_offset_);
                        frm.DetectAllDrive();
                        frm.ReadGameOnWii();
                    }
                    break;

                case eAction.eACTION_COPY:
                    {
                        WBFS.wbfs_applet_copy(t.source_, t.source_offset_, t.destination_, t.destination_offset_, t.args_);
                        frm.ReadGameOnWii();
                        frm.ReadLocalGame();
                    }
                    break;
            }

            tasks_mutex.WaitOne();
            tasks.RemoveAt(0);
            tasks_mutex.ReleaseMutex();

            frm.IncrementOverall();
            frm.EnableButtons();

            mutex.ReleaseMutex();
        }
    }
}
