#ifdef _WIN32

#include <windows.h>
#include <setupapi.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>

#include "libwbfs.h"

unsigned short sByte(unsigned short n){ return ((n & 0xff) << 8) | ((n & 0xff00) >> 8); }
unsigned int lByte(unsigned int n){ return ((n & 0xff) << 24) | ((n & 0xff00) << 8) | ((n & 0xff0000) >> 8) | ((n & 0xff000000) >> 24); }

static int read_sector(void *_handle, unsigned int lba, unsigned int count, void *buf)
{
	HANDLE handle = (HANDLE)_handle;
	LARGE_INTEGER large;
	DWORD read;
	unsigned long long offset = lba;

	offset *= 512ULL;
	large.QuadPart = offset;

	if (SetFilePointerEx(handle, large, NULL, FILE_BEGIN) == FALSE)
	{
		fprintf(stderr, "\n\n%lld %d %p\n", offset, count, _handle);
		wbfs_error("error seeking in hd sector (read)");
		return 1;
	}

	read = 0;
	if (ReadFile(handle, buf, count * 512ULL, &read, NULL) == FALSE)
	{
		wbfs_error("error reading hd sector");
		return 1;
	}

	return 0;
}

static int write_sector(void *_handle, unsigned int lba, unsigned int count, void *buf)
{
	HANDLE handle = (HANDLE)_handle;
	LARGE_INTEGER large;
	DWORD written;
	unsigned long long offset = lba;

	offset *= 512ULL;
	large.QuadPart = offset;

	if (SetFilePointerEx(handle, large, NULL, FILE_BEGIN) == FALSE)
	{
		wbfs_error("error seeking in hd sector (write)");
		return 1;
	}

	written = 0;
	if (WriteFile(handle, buf, count * 512ULL, &written, NULL) == FALSE)
	{
		DWORD error = GetLastError();
		wbfs_error("error writing hd sector");
		return 1;
	}

	return 0;
}

static int get_capacity(char *fileName, int offset, unsigned int *sector_size, unsigned int *sector_count)
{
	DISK_GEOMETRY dg;
	PARTITION_INFORMATION pi;

	DWORD bytes;
	scopeHANDLE handle = CreateFile(fileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);

	DWORD error = GetLastError();
	if (handle.handle == INVALID_HANDLE_VALUE)
	{
		wbfs_error("could not open drive");
		return 0;
	}

	if (DeviceIoControl(handle.handle, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &dg, sizeof(DISK_GEOMETRY), &bytes, NULL) == FALSE)
	{
		wbfs_error("could not get drive geometry");
		return 0;
	}

	*sector_size = dg.BytesPerSector;

	if(offset == 0)
	{
	if (DeviceIoControl(handle.handle, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &pi, sizeof(PARTITION_INFORMATION), &bytes, NULL) == FALSE)
	{
		wbfs_error("could not get partition info");
		return 0;
	}

	*sector_count = (unsigned int)(pi.PartitionLength.QuadPart / dg.BytesPerSector);
	}
	else
	{
		unsigned char buff[512];
		DWORD fsize;

		// read sector0 of physical drive (partition table)
		if(ReadFile(handle.handle,buff,512,&fsize,NULL) == FALSE)
		{ 
			printf("Cannot read partition table (error %d).\n", GetLastError());
		}

		for(int i=0;i<4;i++)
		{
			// check partition size for entry i
			if(*((unsigned int *)(buff+0x1be+0x0c+0x10*i)) == 0)
				continue;

			if(offset==*((unsigned int *)(buff+0x1be+0x08+0x10*i)))
			{
				*sector_count = (unsigned int)((*((unsigned int *)(buff+0x1be+0x0c+0x10*i))));
			}
		}
	}
	return 1;
}

wbfs_t *wbfs_try_open_hd(char *driveName, int reset)
{
	wbfs_error("no direct harddrive support");
	return 0;
}

wbfs_t *wbfs_try_open_partition(char *partitionLetter, int offset, int reset)
{
	HANDLE handle;
	unsigned int sector_size, sector_count;

	if (!get_capacity(partitionLetter, offset, &sector_size, &sector_count))
	{
		return NULL;
	}

	handle = CreateFile(partitionLetter, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);

	DWORD error = GetLastError();
	if (handle == INVALID_HANDLE_VALUE)
	{
		return NULL;
	}

	LARGE_INTEGER large;
	large.QuadPart = offset * 512ULL;

	if(SetFilePointerEx(handle, large, NULL, FILE_BEGIN) == FALSE)
	{
		return NULL;
	}

	return wbfs_open_partition(read_sector, write_sector, handle, offset, sector_size, sector_count, 0, reset);
}

wbfs_t *wbfs_try_open(char *disc, char *partition, int offset, int reset)
{
	wbfs_t *p = 0;

	if (partition)
	{
		p = wbfs_try_open_partition(partition, offset,reset);
	}

	if (!p && !reset && disc)
	{
		p = 0;
	}
	else if(!p && !reset)
	{
		p = 0;
	}

	return p;
}

#endif

