// Copyright 2009 Kwiirk
// Licensed under the terms of the GNU GPL, version 2
// http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt


//#include <windows.h>
//#include <stdio.h>     /* for printf */
//#include <stdlib.h>    /* for exit */
//#include <sys/stat.h>

#ifdef _WIN32
#include <windows.h>
//#include <direct.h>
//#include <io.h>
#else
//#include <unistd.h>
#endif

//#include <errno.h>
#include "wbfs.h"
#include "wbfs_internal.h"
//#include "tools.h"
#include "../libwbfs/libwbfs.h"

#ifndef GB
#define GB (1024 * 1024 * 1024.)
#define MB (1024 * 1024)
#endif

char pdName[]="\\\\.\\PhysicalDrive0";
void (*progress)(int x, int y) = 0;
void (*text)(const char * text) = 0;

void output_string(const std::string& output)
{
	if(text)
		text(output.c_str());
}

dllport void __stdcall wbfs_init_output_progress(void (*output_progress)(int, int))
{
	progress = output_progress;
}

dllport void __stdcall wbfs_init_output_text(void (*output_text)(const char *))
{
	text = output_text;
}

int read_wii_disc_sector(void *_handle, unsigned int _offset, unsigned int count, void *buf)
{
	HANDLE handle = (HANDLE)_handle;
	LARGE_INTEGER large;
	DWORD read;
	unsigned long long offset = _offset;

	offset <<= 2;
	large.QuadPart = offset;

	if (SetFilePointerEx(handle, large, NULL, FILE_BEGIN) == FALSE)
	{
		wbfs_error("error seeking in disc file");
		return 1;
	}

	read = 0;
	if((ReadFile(handle, buf, count, &read, NULL) == FALSE) || !read)
	{
		wbfs_error("error reading wii disc sector");
		return 1;
	}

	if(read < count)
	{
		wbfs_warning("warning: requested %d, but read only %d bytes (trimmed or bad padded ISO)", count, read);
		wbfs_memset((unsigned char*)buf+read, 0, count-read);
	}

	return 0;
}

int write_wii_disc_sector(void *_handle, unsigned int lba, unsigned int count, void *buf)
{
	HANDLE handle = (HANDLE)_handle;
	LARGE_INTEGER large;
	DWORD written;
	unsigned long long offset = lba;

	offset *= 0x8000;
	large.QuadPart = offset;

	if (SetFilePointerEx(handle, large, NULL, FILE_BEGIN) == FALSE)
	{
		fprintf(stderr,"\n\n%lld %p\n", offset, handle);
		wbfs_error("error seeking in wii disc sector (write)");
		return 1;
	}

	written = 0;
	if (WriteFile(handle, buf, count * 0x8000, &written, NULL) == FALSE)
	{
		wbfs_error("error writing wii disc sector");
		return 1;
	}

	if (written != count * 0x8000)
	{
		wbfs_error("error writing wii disc sector (size mismatch)");
		return 1;
	}

	return 0;
}

dllport void __stdcall wbfs_applet_list_drive(char* mounted_drive, char* drive_list)
{
	char drivePath[8] = "\\\\?\\Z:";

	for(int i = 1; i < 26; ++i)
	{
		HANDLE handle;

		drivePath[4] = 'A' + i;//partitionLetter[0];

		handle = CreateFile(drivePath, 0, 0, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);

		if(handle == INVALID_HANDLE_VALUE)
			continue;

		STORAGE_DEVICE_NUMBER sdn;
		DWORD dwBytesReturned = 0;
		DeviceIoControl(handle, IOCTL_STORAGE_GET_DEVICE_NUMBER, NULL, 0, &sdn, sizeof(sdn), &dwBytesReturned, NULL); 
		sprintf((mounted_drive), "%s%s@PhysicalDrive%d@Part%d%%#%%", mounted_drive, drivePath,sdn.DeviceNumber,sdn.PartitionNumber - 1);

		CloseHandle(handle);
	}

	for(int drvn=1;drvn < 10;drvn++)
	{
		HANDLE hFile;
		unsigned char buff[512];
		unsigned char lbuff[512];
		__int64 offset;
		DWORD fsize;

		// assign drive #
		pdName[sizeof(pdName)-2]='0'+drvn;
		hFile = CreateFile( pdName,               // file to open
			GENERIC_READ,          // open for reading
			FILE_SHARE_READ,       // share for reading
			NULL,                  // default security
			OPEN_EXISTING,         // existing file only
			FILE_ATTRIBUTE_NORMAL,
			NULL);                 // no attr. template

		if(hFile == INVALID_HANDLE_VALUE) 
			continue;

		// read sector0 of physical drive (partition table)
		if(ReadFile(hFile,buff,512,&fsize,NULL) == FALSE)
		{ 
			printf("Cannot read partition table (error %d).\n", GetLastError());
		}
		// check partition table magic number
		else if( *((unsigned short *)(buff+0x1fe)) == 0xaa55 )
		{ 
			for(int i=0;i<4;i++)
			{
				// skip check partition id because WBFS has no unique id
				//if(buff[0x1be+4+0x10*i] == ??)
				//  continue;

				// check partition size for entry i
				if(*((unsigned int *)(buff+0x1be+0x0c+0x10*i)) == 0)
					continue;
				offset=*((unsigned int *)(buff+0x1be+0x08+0x10*i));

				// seek to sector0 of partition (logical drive)
				LARGE_INTEGER large;
				
				large.QuadPart = offset * 512ULL;

				if(SetFilePointerEx(hFile, large,NULL,FILE_BEGIN) == FALSE)
				{ 
					printf("Cannot seek to logical drive %d-%d (error %d).\n",
						drvn,i,GetLastError());
					continue;
				}

				// read sector0 of logical drive
				if(ReadFile(hFile,lbuff,512,&fsize,NULL) == FALSE)
				{ 
					printf("Cannot read logical drive %d-%d (error %d).\n",
						drvn,i,GetLastError());
					continue;
				}

				// check magic for sector0 of logical drive
				if( *((unsigned short *)(lbuff+0x1fe)) == 0xaa55 )
				{
					sprintf((drive_list), "%sUSED@%s@Part%d@LBA%d@size%dsect%%#%%", drive_list,pdName,i,
						*((unsigned int *)(buff+0x1be+0x08+0x10*i)),
						*((unsigned int *)(buff+0x1be+0x0c+0x10*i))
						);
					continue;	// formatted logical drive, don't touch it
				}

				// check wbfs magic "WBFS" for sector0 of logical drive
				if( *((unsigned int *)(lbuff+0)) == 0x53464257 )
				{
					sprintf((drive_list), "%sWBFS@%s@Part%d@LBA%d@size%dsect%%#%%", drive_list,pdName,i,
						*((unsigned int *)(buff+0x1be+0x08+0x10*i)),
						*((unsigned int *)(buff+0x1be+0x0c+0x10*i))
						);	       	       
				}
				else
				{
					sprintf((drive_list), "%sFREE@%s@Part%d@LBA%d@size%dsect%%#%%", drive_list,pdName,i,
						*((unsigned int *)(buff+0x1be+0x08+0x10*i)),
						*((unsigned int *)(buff+0x1be+0x0c+0x10*i))
						);
				}
			}
		}

		CloseHandle(hFile);
	}
}

dllport int __stdcall wbfs_applet_list(char* partition, int offset, char* disc_list)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		return -1;
	}
	scopeHANDLE discHandle = p->callback_data;

	int count = wbfs_count_discs(p);

	if (count == 0)
	{
		return 0;
	}
	else
	{
		int i;
		unsigned int size;
		unsigned char *b = static_cast<unsigned char *>(wbfs_ioalloc(0x100));
		for (i = 0; i < count; i++)
		{
			if (!wbfs_get_disc_info(p, i, b, 0x100, &size))
			{
				sprintf((disc_list), "%s%c%c%c%c%c%c %4dM %s%%#%%", (disc_list), b[0], b[1], b[2], b[3], b[4], b[5], (unsigned int)(size * 4ULL / (MB)), b + 0x20);
			}
		}

		wbfs_iofree(b);
	}
	wbfs_close(p);

	return count;
}

dllport void __stdcall wbfs_applet_info(char* partition, int offset, unsigned long long* used_space, unsigned long long* free_space, unsigned long long* total_space)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	unsigned int count = wbfs_count_usedblocks(p);

	(*total_space) = (unsigned long long)(p->n_wbfs_sec) * p->wbfs_sec_sz;
	(*used_space) = (unsigned long long)(p->n_wbfs_sec-count) * p->wbfs_sec_sz;
	(*free_space) = (unsigned long long)(count) * p->wbfs_sec_sz;

	wbfs_close(p);
}

dllport void __stdcall wbfs_applet_makehbc(char* partition, int offset)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	int count = wbfs_count_discs(p);
	FILE *xml;

	if (count == 0)
	{
		//! TODO ManageError
		//fprintf(stderr,"wbfs is empty\n");
	}
	else
	{
		int i;
		unsigned int size;
		unsigned char *b = static_cast<unsigned char *>(wbfs_ioalloc(0x100));

		for (i = 0; i < count; i++)
		{
			wbfs_get_disc_info(p, i, b, 0x100, &size);

			std::string gameID = boost::lexical_cast<std::string>(b);
			std::string dirname = gameID;
			std::string dolname = gameID + "\\boot.dol";
			std::string pngname = gameID + "\\icon.png";
			std::string xmlname = gameID + "\\meta.xml";

			CreateDirectory(dirname.c_str(), NULL);
			printf("%s\n", dirname);

			CopyFile("boot.dol", dolname.c_str(), FALSE);
			CopyFile("icon.png", pngname.c_str(), FALSE);

			fopen_s(&xml, xmlname.c_str(), "wb");
			fprintf_s(xml,"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n");
			fprintf_s(xml,"<app>\n\t<name>%s</name>\n", b + 0x20);
			fprintf_s(xml,"<short_description>%.2fGB on USB HD</short_description>\n", size * 4ULL / GB);
			fprintf_s(xml,"<long_description>This launches the yal wbfs game loader by Kwiirk for discid %s</long_description>\n", dirname);
			fprintf_s(xml,"</app>");
			fclose(xml);
		}

		wbfs_iofree(b);
	}
	wbfs_close(p);
}

dllport void __stdcall wbfs_applet_init(char* partition, int offset)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, true);
	if (!p)
	{
		wbfs_message("Cannot open WBFS partition");
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	// nothing to do actually..
	// job already done by the reset flag of the wbfs_open_partition
	wbfs_close(p);
	wbfs_message("WBFS partition initialized");
	fprintf(stderr, "wbfs initialized.\n");
}

dllport unsigned int __stdcall wbfs_applet_estimate(char* partition, int offset, char* filename)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		return 0;
	}
	scopeHANDLE discHandle = p->callback_data;

	scopeHANDLE handle = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	if (handle.handle == INVALID_HANDLE_VALUE)
	{
		wbfs_message("Cannot open disc file");
		return 0;
	}

	unsigned int estimated_size = wbfs_estimate_disc(p, read_wii_disc_sector, handle.handle, ONLY_GAME_PARTITION);

	wbfs_close(p);

	return estimated_size;
}

dllport void __stdcall wbfs_applet_add(char* partition, int offset, char* filename)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		wbfs_message("Cannot open WBFS partition");
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	unsigned char discinfo[7];
	wbfs_disc_t *d;

	scopeHANDLE handle = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

	if (handle.handle == INVALID_HANDLE_VALUE)
	{
		wbfs_message("Cannot open disc file");
		fprintf(stderr, "unable to open disc file\n");
	}
	else
	{
		DWORD read = 0;
		ReadFile(handle.handle, discinfo, 6, &read, NULL);

		bool is_iso = true;

		/*if(discinfo[0]=='C' && discinfo[1]=='I' && discinfo[2]=='S' && discinfo[3]=='O')
		{
			LARGE_INTEGER large;
			large.QuadPart=32768ULL;
			SetFilePointerEx(handle.handle, large, NULL, FILE_BEGIN);
			ReadFile(handle.handle, discinfo, 6, &read, NULL);

			is_iso = false;
		}*/

		d = wbfs_open_disc(p, discinfo);

		if (d)
		{
			discinfo[6] = 0;
			wbfs_message("Disc " + boost::lexical_cast<std::string>(discinfo) + " already exist.");
			fprintf(stderr, "%s already in disc...\n", discinfo);
			wbfs_close_disc(d);
		}
		else if(is_iso)
		{
			wbfs_add_disc(p, read_wii_disc_sector, handle.handle, progress, ONLY_GAME_PARTITION, 0);
		}
		/*else
		{
			wbfs_add_ciso_disc(p, read_wii_disc_sector, handle.handle, progress, ONLY_GAME_PARTITION, 0);
		}*/
	}
	wbfs_close(p);
}

dllport void __stdcall wbfs_applet_remove(char* partition, int offset, unsigned char *discID)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		wbfs_message("Cannot open WBFS partition");
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	if(wbfs_rm_disc(p, discID))
	{
		wbfs_message("Couldn't find " + boost::lexical_cast<std::string>(discID));
	}
	else
	{
		printf("Done.\n");
	}
	wbfs_close(p);
}

dllport void __stdcall wbfs_applet_extract(char* partition, int offset, unsigned char* discID, char* directory, char* filename)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		wbfs_message("Cannot open WBFS partition");
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	wbfs_disc_t *d;
	d = wbfs_open_disc(p, discID);

	if (d)
	{
		scopeHANDLE handle;
		std::string isoname;
		int len;

		std::string path = boost::lexical_cast<std::string>(directory);
		isoname = boost::lexical_cast<std::string>(filename);
		len = isoname.length();

		if(len == 0)
		{
			/* get the name of the title to find out the name of the iso */
			isoname = boost::lexical_cast<std::string>(d->header->disc_header_copy + 0x20);
			len = isoname.length();

			// replace silly chars by '_'
			for(int i = 0; i < len; ++i)
			{
				if(isoname[i] == ' ') isoname[i] = '_';
				if(isoname[i] == '|') isoname[i] = '_';
				if(isoname[i] == '/') isoname[i] = '_';
				if(isoname[i] == ':') isoname[i] = '_';
			}
		}

		isoname += ".iso";

		handle = CreateFile((path + isoname).c_str(), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, CREATE_NEW, 0, NULL);

		if (handle.handle == INVALID_HANDLE_VALUE)
		{
			wbfs_message("Cannot write to file " + path + isoname);
			fprintf(stderr, "unable to open disc file (%s) for writing\n", isoname);
		}
		else
		{
			LARGE_INTEGER large;

			fprintf(stderr, "writing to %s\n", isoname);

			large.QuadPart = (d->p->n_wii_sec_per_disc / 2) * 0x8000ULL;
			SetFilePointerEx(handle.handle, large, NULL, FILE_BEGIN);
			SetEndOfFile(handle.handle);

			wbfs_extract_disc(d,write_wii_disc_sector, handle.handle, progress);
		}

		wbfs_close_disc(d);
	}
	else
	{
		wbfs_message("Couldn't find " + boost::lexical_cast<std::string>(discID));
	}
	wbfs_close(p);
}

dllport void __stdcall wbfs_applet_copy(char* src_partition, int src_offset, char* dst_partition, int dst_offset, unsigned char* discID)
{
	wbfs_t *p_src = wbfs_try_open(NULL, src_partition, src_offset, false);
	if (!p_src)
	{
		wbfs_message("Cannot open source WBFS partition");
		return;
	}
	scopeHANDLE srcDiscHandle = p_src->callback_data;

	wbfs_t *p_dst = wbfs_try_open(NULL, dst_partition, dst_offset, false);
	if (!p_dst)
	{
		wbfs_message("Cannot open destination WBFS partition");
		return;
	}
	scopeHANDLE dstDiscHandle = p_dst->callback_data;

	wbfs_disc_t *d;
	d = wbfs_open_disc(p_dst, discID);

	if (d)
	{
		wbfs_message("Disc " + boost::lexical_cast<std::string>(discID) + " already exist.");
		fprintf(stderr, "%s already in disc...\n", discID);
		wbfs_close_disc(d);
	}
	else
	{
		d = wbfs_open_disc(p_src, discID);

		if (d)
		{
			wbfs_copy_disc(d, p_dst, progress);

			wbfs_close_disc(d);
		}
		else
		{
			wbfs_message("Couldn't find " + boost::lexical_cast<std::string>(discID));
			wbfs_error("Couldn't find %s", discID);
		}
	}
	wbfs_close(p_src);
	wbfs_close(p_dst);
}

dllport void __stdcall wbfs_applet_rename(char* partition, int offset, unsigned char* discID, char *newName)
{
	wbfs_t *p = wbfs_try_open(NULL, partition, offset, false);
	if (!p)
	{
		wbfs_message("Cannot open source WBFS partition");
		return;
	}
	scopeHANDLE discHandle = p->callback_data;

	if(wbfs_ren_disc(p, discID, newName))
	{
		wbfs_error("error");
	}
	wbfs_close(p);

	printf("Done.\n");
}
